<?php
/**
 *   Copyright (c) 2018, MonkeyData s.r.o.
 *
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *    * Neither the name of the <organization> nor the
 *   names of its contributors may be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *   @author    MonkeyData
 *   @copyright 2018 MonkeyData s.r.o.
 *   @license   MIT
 */

namespace MonkeyData\Inside\App\Installation;

use Db;
use Tab;
use DbQuery;
use Language;
use Throwable;
use Exception;
use Configuration;
use Monkey\InsideSentry\InsideSentry;
use MonkeyData\Inside\App\ErrorStack\IErrorStack;
use MonkeyData\Inside\CurrentPluginConfiguration;

/**
 * Class AFoundationHandler
 * @package MonkeyData\Inside\App\Installation
 */
abstract class AFoundationHandle
{
    const ICON_PATH = '/views/img';

    /**
     * @var int
     */
    private $pluginsCount;

    /**
     * @var int
     */
    private $parentTabId;

    /**
     * @var PrestaShopSecurityHandler
     */
    private $securityHandler;

    /**
     * @var IErrorStack
     */
    private $errorStack;

    /**
     * AFoundationHandle constructor.
     * @param PrestaShopSecurityHandler $securityHandler
     * @param IErrorStack $errorStack
     */
    public function __construct(PrestaShopSecurityHandler $securityHandler, IErrorStack $errorStack)
    {
        $pluginsCount = Configuration::get(CurrentPluginConfiguration::CONFIGURATION_PLUGINS_COUNT);
        if (empty($pluginsCount)) {
            $pluginsCount = 0;
        }
        $this->setPluginsCount($pluginsCount)->setSecurityHandler($securityHandler)->setErrorStack($errorStack);
    }

    /**
     * @return bool
     */
    public function install()
    {
        if ($this->getPluginsCount() > 0) {
            return true;
        }
        if (version_compare('1.6.0', _PS_VERSION_, '>=')) {
            $engine = _MYSQL_ENGINE_;
        } else {
            $engine = Db::getInstance()->getBestEngine();
        }
        $tablesToCreate = $this->getSecurityHandler()->getTableCreates(_DB_NAME_, $engine);
        foreach ($tablesToCreate as $tableToCreate) {
            if (Db::getInstance()->execute($tableToCreate) === false) {
                $this->getErrorStack()->add('Failed to create mandatory database table');
                InsideSentry::error('Unable to create table via ' . var_export($tableToCreate, true));
                return false;
            }
        }

        $this->setupApiAccessToken();

        Configuration::updateValue(CurrentPluginConfiguration::CONFIGURATION_PLUGINS_COUNT, 0);
        return true;
    }

    /**
     * Setups Api Access Token
     */
    public function setupApiAccessToken()
    {
        $apiAccessToken = $this->generateToken();
        $this->getSecurityHandler()->storeTokenToLocalStorage($apiAccessToken);
    }

    /**
     * @return string
     */
    private function generateToken()
    {
        return hash('sha256', uniqid(Configuration::get('PS_SHOP_NAME'), true));
    }


    /**
     * @return int|false
     */
    final protected function getParentTabId()
    {
        if ($this->parentTabId === null) {
            $this->parentTabId = Tab::getIdFromClassName(CurrentPluginConfiguration::PARENT_TAB_NAME);
            if ($this->parentTabId === false) {
                try {
                    $this->parentTabId = $this->installParentTab()->id;
                } catch (Exception $exception) {
                    $this->getErrorStack()->add('Failed to create Tab ' . CurrentPluginConfiguration::PARENT_TAB_NAME);
                    return null;
                } catch (Throwable $exception) {
                    $this->getErrorStack()->add('Failed to create Tab ' . CurrentPluginConfiguration::PARENT_TAB_NAME);
                    return null;
                }
            }
        }
        return $this->parentTabId;
    }

    /**
     * @return Tab
     */
    private function installParentTab()
    {
        $tab = new Tab();
        foreach (Language::getLanguages(false) as $lang) {
            $tab->name[(int) $lang['id_lang']] = CurrentPluginConfiguration::PARENT_TAB_NAME;
        }
        $tab->module = null;
        $tab->id_parent = 0; // Root tab
        $tab->class_name = CurrentPluginConfiguration::PARENT_TAB_NAME;
        $tab->active = 1;
        $tab->add();

        $data = array(
            'module' => $this->getParentTabIconPath(CurrentPluginConfiguration::PLUGIN_NAME)
        );
        Db::getInstance()->update('tab', $data, "`id_tab` = {$tab->id}");

        return $tab;
    }


    /**
     * @return bool
     */
    public function uninstall()
    {
        $this->handleParentTabOwnership();

        if ($this->getPluginsCount() > 0) {
            return true;
        }

        $this->uninstallParentTab();

        $tablesToDrop = $this->getSecurityHandler()->getTableDrops(_DB_NAME_);
        foreach ($tablesToDrop as $tableToDrop) {
            if (Db::getInstance()->execute($tableToDrop) === false) {
                $this->getErrorStack()->add('Failed to drop table');
            }
        }
        Configuration::deleteByName(CurrentPluginConfiguration::CONFIGURATION_PLUGINS_COUNT);
        return true;
    }

    /**
     * @return bool
     */
    private function uninstallParentTab()
    {
        try {
            $parentTabId = $this->loadTabId(
                CurrentPluginConfiguration::PARENT_TAB_NAME,
                $this->getParentTabIconPath(CurrentPluginConfiguration::PLUGIN_NAME)
            );
            if ($parentTabId === null) {
                $this->getErrorStack()->add('Supposed parent not found');
                return true;
            }

            $tab = new \Tab();

            foreach (Language::getLanguages(false) as $lang) {
                $tab->name[(int) $lang['id_lang']] = CurrentPluginConfiguration::PARENT_TAB_NAME;
            }
            $tab->module = CurrentPluginConfiguration::PLUGIN_NAME;
            $tab->class_name = CurrentPluginConfiguration::PARENT_TAB_NAME;
            $tab->id = $parentTabId;
            $tab->delete();
        } catch (Exception $e) {
            $this->getErrorStack()->add('Failed to delete Tab ' . CurrentPluginConfiguration::PARENT_TAB_NAME);
            return false;
        } catch (Throwable $e) {
            $this->getErrorStack()->add('Failed to delete Tab ' . CurrentPluginConfiguration::PARENT_TAB_NAME);
            return false;
        }

        return true;
    }

    /**
     * @return void
     */
    private function handleParentTabOwnership()
    {
        $parentTabId = $this->loadTabId(
            CurrentPluginConfiguration::PARENT_TAB_NAME,
            $this->getParentTabIconPath(CurrentPluginConfiguration::PLUGIN_NAME)
        );
        if ($parentTabId === null) {
            return;
        }
        $childTabIds = $this->getTabChildIds((int)$parentTabId);
        if ($childTabIds === null) {
            return;
        }

        $currentTabId = $this->loadTabId(
            CurrentPluginConfiguration::PLUGIN_TAB_CONTROLLER,
            CurrentPluginConfiguration::PLUGIN_NAME
        );

        foreach ($childTabIds as $childTabId) {
            if ($childTabId === $currentTabId) {
                continue;
            }
            if ($this->passParentTabOwnership((int)$parentTabId, (int)$childTabId) === false) {
                $this->getErrorStack()->add('Error passing child icon to parent');
            }
            break;
        }
    }

    /**
     * @param $parentTabId
     * @param $childTabId
     * @return bool
     */
    private function passParentTabOwnership($parentTabId, $childTabId)
    {
        $builder = new DbQuery();
        $builder = $builder->select('module')
            ->from('tab')
            ->where("`id_tab` = {$childTabId}");

        $result = Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue($builder, false);

        if ($result === false) {
            return false;
        }

        $data = array(
            'module' => $this->getParentTabIconPath($result)
        );

        Db::getInstance()->update('tab', $data, "`id_tab` = {$parentTabId}", 1);

        return true;
    }

    /**
     * @param string $className
     * @param string $module
     * @return null|string
     */
    protected function loadTabId($className, $module)
    {
        $builder = new DbQuery();
        $builder = $builder->select('id_tab')
            ->from('tab')
            ->where("`class_name` = '{$className}' && `module` = '{$module}'");

        $result = Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue($builder, false);

        if ($result === false) {
            return null;
        }
        return (int)$result;
    }

    /**
     * @param $parentTabId
     * @return array|null
     */
    private function getTabChildIds($parentTabId)
    {
        $builder = new DbQuery();
        $builder = $builder->select('id_tab')
            ->from('tab')
            ->where("`id_parent` = {$parentTabId}");

        $result = Db::getInstance(_PS_USE_SQL_SLAVE_)->executeS($builder);

        if (empty($result)) {
            return null;
        }
        $ids = array();

        foreach ($result as $tabId) {
            $ids[] = $tabId['id_tab'];
        }
        return $ids;
    }

    /**
     * @param $moduleName
     * @return string
     */
    private function getParentTabIconPath($moduleName)
    {
        return  $moduleName . static::ICON_PATH;
    }

    /**
     * @return int
     */
    final protected function getPluginsCount()
    {
        return $this->pluginsCount;
    }

    /**
     * @param int $pluginsCount
     * @return $this
     */
    private function setPluginsCount($pluginsCount)
    {
        $this->pluginsCount = $pluginsCount;
        return $this;
    }

    /**
     * @return void
     */
    final protected function incrementPluginsCount()
    {
        $this->pluginsCount++;
        Configuration::updateValue(
            CurrentPluginConfiguration::CONFIGURATION_PLUGINS_COUNT,
            $this->getPluginsCount()
        );
    }

    /**
     * @return void
     */
    final protected function decrementPluginsCount()
    {
        if ($this->pluginsCount <= 0) {
            return;
        }
        $this->pluginsCount--;
        Configuration::updateValue(
            CurrentPluginConfiguration::CONFIGURATION_PLUGINS_COUNT,
            $this->getPluginsCount()
        );
    }

    /**
     * @return PrestaShopSecurityHandler
     */
    protected function getSecurityHandler()
    {
        return $this->securityHandler;
    }

    /**
     * @param PrestaShopSecurityHandler $securityHandler
     * @return $this
     */
    private function setSecurityHandler(PrestaShopSecurityHandler $securityHandler)
    {
        $this->securityHandler = $securityHandler;
        return $this;
    }

    /**
     * @return IErrorStack
     */
    protected function getErrorStack()
    {
        return $this->errorStack;
    }

    /**
     * @param IErrorStack $errorStack
     * @return $this
     */
    private function setErrorStack(IErrorStack $errorStack)
    {
        $this->errorStack = $errorStack;
        return $this;
    }
}
