<?php
/**
 *   Copyright (c) 2018, MonkeyData s.r.o.
 *
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *    * Neither the name of the <organization> nor the
 *   names of its contributors may be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *   @author    MonkeyData
 *   @copyright 2018 MonkeyData s.r.o.
 *   @license   MIT
 */

namespace MonkeyData\Inside\App\Installation;

use Monkey\InsideBootJs\Exceptions\InstallException;
use Tab;
use Language;
use Exception;
use Throwable;
use Monkey\InsideSentry\InsideSentry;
use MonkeyData\Inside\App\PrestaShopConfig;
use MonkeyData\Inside\App\PrestaShopModel;
use MonkeyData\Inside\CurrentPluginConfiguration;
use MonkeyData\Inside\App\PrestaShopRemoteInstallController;

/**
 * Class PluginInstallationHandle
 * @package MonkeyData\Inside\App\Installation
 */
class PluginInstallationHandle extends AFoundationHandle
{

    /**
     * @return bool
     */
    public function install()
    {
        if (parent::install() === false) {
            return false;
        }

        if ($this->getPluginsCount() > 0) {
            try {
                $this->remoteInstallation();
            } catch (Exception $exception) {
                InsideSentry::warning($exception);
            } catch (Throwable $exception) {
                InsideSentry::error($exception);
            }
        }

        $this->incrementPluginsCount();

        if ($this->installTab() === false) {
            return false;
        }

        return true;
    }

    /**
     * @return bool
     */
    private function installTab()
    {
        try {
            $parentTabId = $this->getParentTabId();
            $tab = new Tab();
            foreach (Language::getLanguages(false) as $lang) {
                $tab->name[(int) $lang['id_lang']] = CurrentPluginConfiguration::PLUGIN_TAB_NAME;
            }
            $tab->module = CurrentPluginConfiguration::PLUGIN_NAME;
            $tab->id_parent = $parentTabId;
            $tab->class_name = CurrentPluginConfiguration::PLUGIN_TAB_CONTROLLER;
            if (property_exists($tab, 'icon')) {
                $tab->icon = 'settings_applications';
            }
            $tab->active = 1;
            $tab->add();
        } catch (Exception $e) {
            $this->getErrorStack()->add('Failed to create Tab ' . CurrentPluginConfiguration::PLUGIN_TAB_NAME);
            return false;
        } catch (Throwable $e) {
            $this->getErrorStack()->add('Failed to create Tab ' . CurrentPluginConfiguration::PLUGIN_TAB_NAME);
            return false;
        }
        return true;
    }

    /**
     * @return bool
     */
    public function uninstall()
    {
        try {
            $this->uninstallNotifyMonkeyData();
        } catch (Exception $exception) {
            InsideSentry::warning($exception);
        } catch (Throwable $exception) {
            InsideSentry::error($exception);
        }

        $result = true;

        if ($this->uninstallTab() === false) {
            $result = false;
        }

        $this->decrementPluginsCount();

        if (parent::uninstall() === false) {
            $result = false;
        }

        return $result;
    }

    /**
     * @return bool
     */
    private function uninstallTab()
    {
        try {
            $tab = new Tab();

            foreach (Language::getLanguages(false) as $lang) {
                $tab->name[(int) $lang['id_lang']] = CurrentPluginConfiguration::PLUGIN_TAB_NAME;
            }
            $tab->module = CurrentPluginConfiguration::PLUGIN_NAME;
            $tab->class_name = CurrentPluginConfiguration::PLUGIN_TAB_CONTROLLER;
            $tab->id = $this->loadTabId($tab->class_name, $tab->module);
            $tab->delete();
        } catch (Exception $e) {
            $this->getErrorStack()->add('Failed to delete Tab ' . CurrentPluginConfiguration::PLUGIN_TAB_NAME);
            return false;
        } catch (Throwable $e) {
            $this->getErrorStack()->add('Failed to delete Tab ' . CurrentPluginConfiguration::PLUGIN_TAB_NAME);
            return false;
        }
        return true;
    }

    /**
     * Notifies MonkeyData about installation
     */
    private function remoteInstallation()
    {
        $config = new PrestaShopConfig();
        $model = new PrestaShopModel();
        $controller = new PrestaShopRemoteInstallController($model, $config);
        $controller->install();
    }

    /**
     * Notifies MonkeyData about uninstallation
     */
    private function uninstallNotifyMonkeyData()
    {
        $config = new PrestaShopConfig();
        $model = new PrestaShopModel();
        $controller = new PrestaShopRemoteInstallController($model, $config);
        $controller->uninstall();
    }
}
