<?php
/**
 *   Copyright (c) 2018, MonkeyData s.r.o.
 *
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *    * Neither the name of the <organization> nor the
 *   names of its contributors may be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *   @author    MonkeyData
 *   @copyright 2018 MonkeyData s.r.o.
 *   @license   MIT
 */

namespace MonkeyData\Inside\App;

use Monkey\InsideBootJs\Model\AModel;
use Monkey\InsidePlatformData\Platform\PrestaShopConnectionStructure;
use Configuration;
use Db;
use Monkey\Oauth2Token\AccessToken;
use Monkey\Oauth2Token\Exceptions\Oauth2CreationException;
use Monkey\Oauth2Token\Interfaces\IAccessToken;
use Tools;
use DbQuery;
use Monkey\PluginsPrestashop17XApi\LocalhostSecurityHandler;
use MonkeyData\Inside\App\Exceptions\PluginConfigurationQueryException;
use MonkeyData\Inside\CurrentPluginConfiguration;

/**
 * Class PrestaShopModel
 */
class PrestaShopModel extends AModel
{

    const TYPE_SECRET_TOKEN = 'secret_token';
    const TYPE_IFRAME_TOKEN = 'iframe_token';

    /**
     * @return null|IAccessToken
     */
    protected function loadSecretToken()
    {
        try {
            $serialized = $this->getRowData(static::TYPE_SECRET_TOKEN);
        } catch (PluginConfigurationQueryException $e) {
            return null;
        }
        if (empty($serialized)) {
            return null;
        }
        $json = json_decode($serialized);
        if (empty($json) || !($json instanceof \stdClass)) {
            return null;
        }

        try {
            return AccessToken::make($json);
        } catch (Oauth2CreationException $exception) {
            return null;
        }
    }

    /**
     * @param IAccessToken $secretToken
     * @throws \PrestaShopDatabaseException
     */
    protected function saveSecretToken(IAccessToken $secretToken)
    {
        $serialized = $secretToken->toJSON();
        $this->insertOrUpdateRowData(static::TYPE_SECRET_TOKEN, $serialized);
    }

    /**
     *
     */
    public function deleteSecretToken()
    {
        $where = "type = ".static::TYPE_SECRET_TOKEN;
        Db::getInstance()->delete(LocalhostSecurityHandler::PLUGINS_CONFIG_TABLE, $where);
    }


    /**
     * @return IAccessToken|null
     */
    protected function loadIframeToken()
    {
        try {
            $serialized = $this->getRowData(static::TYPE_IFRAME_TOKEN);
        } catch (PluginConfigurationQueryException $e) {
            return null;
        }
        if (empty($serialized)) {
            return null;
        }
        $json = json_decode($serialized);
        if (empty($json) || !($json instanceof \stdClass)) {
            return null;
        }

        try {
            return AccessToken::make($json);
        } catch (Oauth2CreationException $exception) {
            return null;
        }
    }

    /**
     * @param IAccessToken $iframeToken
     * @throws \PrestaShopDatabaseException
     */
    protected function saveIframeToken(IAccessToken $iframeToken)
    {
        $serialized = $iframeToken->toJSON();

        $this->insertOrUpdateRowData(static::TYPE_IFRAME_TOKEN, $serialized);
    }

    /**
     * @return PrestaShopConnectionStructure
     * @throws PluginConfigurationQueryException
     */
    protected function loadConnectionData()
    {
        $apiAccessToken = $this->getApiAccessToken();
        if ($apiAccessToken === null) {
            throw new PluginConfigurationQueryException('Unable to load API Access token');
        }

        $identity = new PrestaShopConnectionStructure();

        $identity->setEmail(Configuration::get('PS_SHOP_EMAIL'))
            ->setFirstname('Firstname')
            ->setLastname('Lastname')
            ->setStoreIdentityHash($this->loadStoreIdentityHash())
            ->setApiAccessToken($apiAccessToken)
            ->setShopIds($this->getShopIds())
            ->setShopGroupIds($this->getShopGroupIds())
            ->setApiVersion(CurrentPluginConfiguration::API_VERSION)
            ->setApiUrl($this->composeApiUrl())
            ->setLanguage($this->getLanguage())
            ->setStoreUrl(Tools::getShopDomainSsl(true))
            ->setCurrency($this->getCurrency())
            ->setCountry($this->getCountry())
            ->setTimezone($this->getTimezone());

        return $identity;
    }

    /**
     * @return string
     */
    private function getLanguage()
    {
        return Configuration::get('PS_LOCALE_LANGUAGE');
    }

    /**
     * @return null|string
     */
    private function getCurrency()
    {
        $currencyId = (int) Configuration::get('PS_CURRENCY_DEFAULT');
        $query = new DbQuery();
        $query->select('iso_code')
            ->from('currency')
            ->where("id_currency = {$currencyId}");

        try {
            return Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue($query->build());
        } catch (\PrestaShopException $exception) {
            return null;
        }
    }

    /**
     * @return string
     */
    private function getCountry()
    {
        return Configuration::get('PS_LOCALE_COUNTRY');
    }

    /**
     * @return string
     */
    private function getTimezone()
    {
        return Configuration::get('PS_TIMEZONE');
    }

    /**
     * @return null|string
     * @throws PluginConfigurationQueryException
     */
    public function getApiAccessToken()
    {
        return $this->getRowData(LocalhostSecurityHandler::TYPE_API_ACCESS_TOKEN);
    }

    /**
     * @return string
     */
    protected function loadStoreIdentityHash()
    {
        $shopName = Configuration::get('PS_SHOP_NAME');
        return hash('sha256', $shopName);
    }

    /**
     * @return string
     */
    private function composeApiUrl()
    {
        return Tools::getShopDomainSsl(true) . _MODULE_DIR_ . CurrentPluginConfiguration::PLUGIN_NAME
            . '/' . 'api.php';
    }

    /**
     * @return array
     */
    private function getShopIds()
    {
        $query = new DbQuery();
        $query->select('id_shop, id_shop_group, name')->from('shop');
        $result = Db::getInstance(_PS_USE_SQL_SLAVE_)->executeS($query->build());
        if (empty($result)) {
            return array();
        }
        return $result;
    }

    /**
     * @return array
     */
    private function getShopGroupIds()
    {
        $query = new DbQuery();
        $query->select('id_shop_group, name')->from('shop_group');
        $result = Db::getInstance(_PS_USE_SQL_SLAVE_)->executeS($query->build());
        if (empty($result)) {
            return array();
        }
        return $result;
    }

    /**
     * @param $type
     * @param $data
     * @throws \PrestaShopDatabaseException
     */
    private function insertOrUpdateRowData($type, $data)
    {
        $exists = true;
        try {
            $result = $this->getRowData($type);
            if ($result === false) {
                $exists = false;
            }
        } catch (PluginConfigurationQueryException $e) {
            $exists = false;
        }

        $data = array('type' => $type, 'data' => $data);

        if ($exists) {
            Db::getInstance()->update(
                LocalhostSecurityHandler::PLUGINS_CONFIG_TABLE,
                $data,
                "type = '{$type}'"
            );
        } else {
            Db::getInstance()->insert(LocalhostSecurityHandler::PLUGINS_CONFIG_TABLE, $data);
        }
    }

    /**
     * @param $type
     * @return false|null|string
     * @throws PluginConfigurationQueryException
     */
    private function getRowData($type)
    {
        $query = new DbQuery();
        $query->select('data')
            ->from(LocalhostSecurityHandler::PLUGINS_CONFIG_TABLE)
            ->where("type = '{$type}'");

        try {
            $result = Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue($query->build());
        } catch (\PrestaShopException $exception) {
            throw new PluginConfigurationQueryException('No result provided when querying type: ' . $type);
        }
        return $result;
    }
}
