<?php

namespace Monkey\Curl\Base;

use Exception;
use Monkey\Curl\Base\Response\Response;
use Monkey\Curl\Exceptions\HttpCurlCommunicationException;
use Monkey\Curl\Exceptions\HttpCurlResponseDecodingException;
use Monkey\Curl\Exceptions\HttpCurlEndpointNotFoundException;
use Monkey\Curl\Exceptions\HttpCurlEndpointNotImplementedException;
use Monkey\Curl\Exceptions\HttpCurlException;
use Monkey\Curl\Exceptions\HttpCurlUnauthorizedException;
use Monkey\Curl\Interfaces\ICurl;
use stdClass;
use Throwable;

/**
 * Class ACurl
 * @package Monkey\Curl\Base
 */
abstract class ACurl implements ICurl {

    /**
     * ACurl constructor.
     */
    public function __construct() {
    }

    /**
     * @param string $url
     * @param array $parameters
     * @param string $method
     *
     * @return Response
     * @throws HttpCurlCommunicationException
     * @throws HttpCurlException
     */
    final public function call($url, array $parameters, $method = self::METHOD_POST) {

        try {
            $curl = curl_init();

            $options = array(
                CURLOPT_POST => ($method === static::METHOD_POST),
                CURLOPT_HEADER => 0,
                CURLOPT_FRESH_CONNECT => 1,
                CURLOPT_RETURNTRANSFER => 1,
                CURLOPT_FORBID_REUSE => 1,
                CURLOPT_TIMEOUT => 3,
                CURLOPT_URL => $url,
                CURLOPT_POSTFIELDS => $parameters,
                CURLOPT_FOLLOWLOCATION => true
            );

            curl_setopt_array($curl, $options);

            $response = curl_exec($curl);

            $error = curl_error($curl);
            $errorNumber = curl_errno($curl);
            $curlinfo = curl_getinfo($curl);
            curl_close($curl);

            switch ($curlinfo["http_code"]) {
                case 401:
                    throw new HttpCurlUnauthorizedException($url);
                    break;

                case 404:
                    throw new HttpCurlEndpointNotFoundException($url);
                    break;

                case 417:
                    throw new HttpCurlEndpointNotImplementedException($url);
                    break;

                default:
                    break;
            }

            if ($error !== '') {
                throw new HttpCurlCommunicationException($error);
            }

            $object = json_decode($response);

            if ($object === null) {
                throw new HttpCurlResponseDecodingException($url, $parameters);
            }

            if (isset($object->error)) {
                if (isset($object->error_description)) {
                    $message = $object->error_description;
                } else {
                    $message = $object->error;
                }
                if (empty($message)) {
                    $message = (string)$response;
                }
                throw new HttpCurlException($message);
            }
            return $this->parseResponse($object, $curlinfo);
        } catch (HttpCurlException $exception) {
            throw $exception;

        } catch (Exception $exception) {
            throw new HttpCurlCommunicationException("Unknown communication error.", 0, $exception);
        } catch (Throwable $exception) {
            throw new HttpCurlCommunicationException("Unknown communication error.", 0, $exception);
        }
    }

    /**
     * @param stdClass $object
     * @param array $curlInfo
     * @return Response
     */
    abstract protected function parseResponse(stdClass $object, array $curlInfo);


}

