<?php

namespace Monkey\ImportEshopApi\Base;

use Monkey\ImportEshopApi\Exceptions\DataMiningFailedException;
use Monkey\ImportEshopDataObjects\Interfaces\IRequest;
use Monkey\ImportEshopDataObjects\Interfaces\IResponse;

/**
 * Class ADataMiner
 * @package Monkey\ImportEshopApi\Base
 */
abstract class ADataMiner
{


    const TABLE_SCHEMA_UNAVAILABLE = "Table schema for this endpoint is unavailable.";

    /**
     * @var string
     */
    private $endpoint;

    /**
     * @var IResponse
     */
    private $response;

    /**
     * @var IRequest
     */
    private $request;

    /**
     * @var ADefaultEshopSetting
     */
    private $defaultEshopSetting;

    /**
     * ADataMiner constructor.
     * @param IRequest $request
     * @param IResponse $response
     * @param string $endpoint
     * @param ADefaultEshopSetting $defaultEshopSetting
     */
    public function __construct(IRequest $request,IResponse $response, $endpoint, $defaultEshopSetting){
        $this->setRequest($request);
        $this->setResponse($response);
        $this->setEndpoint($endpoint);
        $this->setDefaultEshopSetting($defaultEshopSetting);
    }

    private function setupDebugResponse(){
        $this->getResponse()->setPhpVersion(phpversion());
        $this->getResponse()->setQuery($this->getComposedQuery());
        $this->getResponse()->setEshopPlatform(Config::ESHOP_PLATFORM);
        $this->getResponse()->setEshopPlatformVersion(Config::ESHOP_PLATFORM_VERSION);
        $this->getResponse()->setEshopPluginVersion(Config::ESHOP_PLUGIN_VERSION);
        $this->getResponse()->setImportEshopApiVersion(Config::IMPORT_ESHOP_API_VERSION);
        $this->getResponse()->setImportEshopDataObjectsVersion(Config::IMPORT_ESHOP_DATA_OBJECTS_VERSION);
        try{
            $this->getResponse()->setSchema($this->getTableSchema());
        }catch (DataMiningFailedException $exception){
            $this->getResponse()->setSchema(self::TABLE_SCHEMA_UNAVAILABLE);
        }

    }

    public function prepareResponse(){

        if($this->getRequest()->isDebug()){
            $this->setupDebugResponse();
            $this->getRequest()->setLimit(1);//example for debug
        }
        try {
            $this->setupResponseData();
        }catch(DataMiningFailedException $exception){
            $this->setupDebugResponse();
            $this->getResponse()->setStatusCode($exception->getCode());
            $this->getResponse()->setMessage($exception->getMessage().PHP_EOL.$exception->getPrevious()->getMessage());
        }

    }


    /**
     * @return string
     */
    abstract protected function getComposedQuery();

    /**
     * @return array()|null
     * @throws DataMiningFailedException
     */
    abstract protected function getTableSchema();

    /**
     * @return void
     */
    abstract protected function setupResponseData();

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    abstract protected function getDefaultSelectAttributes($builder = null);
    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    abstract protected function getDefaultTable($builder = null);
    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    abstract protected function getDefaultCondition($builder = null);
    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    abstract protected function getDefaultLimits($builder = null);

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    abstract protected function getDifferentAttributes($builder = null);
    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    abstract protected function getDifferentTable($builder = null);
    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    abstract protected function getDifferentConditions($builder = null);

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getSelectAttributes($builder = null){

        if(
            !is_null($this->getRequest()->getDifferences())
            AND
            !is_null($this->getRequest()->getDifferences()->getColumns())
        ){
            if(!is_null($attributes = $this->getDifferentAttributes($builder))){
                return $attributes;
            }
        }
        return $this->getDefaultSelectAttributes($builder);
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getTable($builder = null){

        if(
            !is_null($this->getRequest()->getDifferences())
            AND
            !is_null($this->getRequest()->getDifferences()->getTableAliases())
        ){
            if(!is_null($tables = $this->getDifferentTable($builder))){
                return $tables;
            }
        }
        return $this->getDefaultTable($builder);
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getConditions($builder = null){
        if(
            !is_null($this->getRequest()->getDifferences())
            AND
            !is_null($this->getRequest()->getDifferences()->getAdditionalConditions())
        ){
            if(!is_null($additionalConditions = $this->getDifferentConditions($builder))){
                return $additionalConditions;
            }
        }

        return $this->getDefaultCondition($builder);
    }

    protected function getLimits($builder = null){
        return $this->getDefaultLimits($builder);
    }

    /**
     * @return int
     */
    protected function getOffset(){
        return ($this->getRequest()->getPage() - 1) * $this->getRequest()->getLimit();
    }

    /**
     * @return int
     */
    protected function getResultsCount(){
        return $this->getRequest()->getLimit();
    }

    //--------------  properties set/get -------------

    /**
     * @return string
     */
    public function getEndpoint(){
        return $this->endpoint;
    }

    /**
     * @param string $endpoint
     * @return ADataMiner
     */
    private function setEndpoint($endpoint){
        $this->endpoint = $endpoint;
        return $this;
    }

    /**
     * @return IRequest
     */
    public function getRequest(){
        return $this->request;
    }

    /**
     * @param IRequest $request
     * @return static
     */
    private function setRequest(IRequest $request){
        $this->request = $request;
        return $this;
    }

    /**
     * @return IResponse
     */
    protected function getResponse(){
        return $this->response;
    }

    /**
     * @param IResponse $response
     * @return ADataMiner
     */
    private function setResponse($response){
        $this->response = $response;
        return $this;
    }

    /**
     * @return ADefaultEshopSetting
     */
    protected function getDefaultEshopSetting(){
        return $this->defaultEshopSetting;
    }

    /**
     * @param ADefaultEshopSetting $defaultEshopSetting
     * @return ADataMiner
     */
    private function setDefaultEshopSetting($defaultEshopSetting){
        $this->defaultEshopSetting = $defaultEshopSetting;
        return $this;
    }


}