<?php

namespace Monkey\ImportEshopApi\Base;

use Monkey\ImportEshopApi\Endpoint\Test;
use Monkey\ImportEshopApi\Exceptions\MissingDefaultEshopSettingClassException;
use Monkey\ImportEshopApi\Exceptions\MissingRequestException;
use Monkey\ImportEshopApi\Interfaces\IImportEshopApi;
use Monkey\ImportEshopApi\Interfaces\ISecurityHandler;
use Monkey\ImportEshopApi\Request\RequestBuilder;
use Monkey\ImportEshopApi\Response\ResponseBuilder;
use Monkey\ImportEshopDataObjects\Request\FakeRequest;
use Monkey\ImportEshopDataObjects\Response\ErrorResponse;
use Monkey\ImportEshopDataObjects\Response\FakeResponse;
use NilPortugues\Serializer\Serializer;

/**
 * Class AImportEshopApi
 * @package Monkey\ImportEshopApi\Base
 */
abstract class AImportEshopApi implements IImportEshopApi {

    const PARAMETER_ENDPOINT = "endpoint";
    const PARAMETER_REQUEST = "request";


    /**
     * @var Serializer
     */
    private $serializer;
    /**
     * @var ISecurityHandler
     */
    private $securityHandler;

    /**
     * @var RequestBuilder
     */
    private $requestBuilder;

    /**
     * @var ResponseBuilder
     */
    private $responseBuilder;

    /**
     * @var MinerFactory
     */
    private $minerFactory;

    /**
     * @var FakeRequest
     */
    private $fakeRequest;
    /**
     * @var FakeResponse
     */
    private $fakeResponse;

    /**
     * @var ADefaultEshopSetting
     */
    private $defaultEshopSetting;

    /**
     * @var array
     */
    private $errors = array();

    /**
     * AImportEshopApi constructor.
     *
     * @param ISecurityHandler $securityHandler
     * @param Serializer $serializer
     */
    public function __construct($securityHandler,$serializer) {
        $this->setSecurityHandler($securityHandler);
        $this->setSerializer($serializer);
    }


    /**
     * @return string
     */
    abstract protected function handleEndpointInput();

    /**
     * @return string
     */
    abstract protected function handleRequestInput();

    abstract protected function registerEndpoints();

    abstract protected function registerDefaultEshopSetting();

    /**
     *
     */
    public function start() {
//        $nearray = "";
//        $x = $nearray["asd"];
//        foreach ($nearray as $x){}
//        $array = array();
//        $x = $array['asd'];
//        $nula = 100 / 0;

        try {
            $this->registerDefaultEshopSetting();
            $this->registerEndpoint(new Test($this->getFakeRequest(),$this->getFakeResponse(), $this->getDefaultEshopSetting()));
            $this->registerEndpoints();
            $endpoint = $this->handleEndpointInput();
            $requestString = $this->handleRequestInput();
        }catch (\Exception $exception){
            $this->makeErrorResponse($exception);
            exit;
        }catch (\Throwable $exception) {
            $this->makeErrorResponse($exception);
            exit;
        }

        if(is_null($requestString)){
            $this->makeErrorResponse(new MissingRequestException());
        }

        try {
            $request = $this->getRequestBuilder()->build($requestString);
            $this->getSecurityHandler()->checkToken($request->getHash());
        }catch (\Exception $exception){
            $this->makeErrorResponse($exception);
            exit;
        } catch (\Throwable $exception) {
            $this->makeErrorResponse($exception);
            exit;
        }

        $this->getResponseBuilder()->setupResponse($request);

        $response = $this->getResponseBuilder()->getResponse();

        try{
            $currentEndpointDataMiner = $this->getMinerFactory()->getMiner($endpoint,$request,$response, $this->getDefaultEshopSetting());
            $currentEndpointDataMiner->prepareResponse();
        }catch(\Exception $exception){
            $this->makeErrorResponse($exception);
        } catch (\Throwable $exception) {
            $this->makeErrorResponse($exception);
        }

        $this->getResponseBuilder()->addErrors($this->getErrors());
        $this->getResponseBuilder()->render();
    }


    /**
     * @param \Exception $exception
     */
    private function makeErrorResponse($exception){

        $errorResponse = new ErrorResponse();
        $errorResponse->setStatusCode($exception->getCode());
        $errorResponse->setMessage($exception->getMessage());

        $error = $errorResponse->addNewEntity();
        $error->setExceptionName(get_class($exception));
        $error->setMessage($exception->getMessage());
        $error->setStackTrace($exception->getTrace());
        $error->setFile($exception->getFile());
        $error->setLine($exception->getLine());

        $this->getResponseBuilder()->setResponse($errorResponse);
        $this->getResponseBuilder()->render();
    }

    /**
     * @param $error_number
     * @param $error_message
     * @param $error_file
     * @param $error_line
     * @param $error_context
     */
    public function errorHandle($error_number, $error_message, $error_file, $error_line, $error_context){
        $errorArray = array("message"=>$error_message,"file"=>$error_file,"line"=>$error_line,
        //    "context"=>$error_context
        );
        switch ($error_number){
            case E_NOTICE:
            case E_WARNING:
            case E_DEPRECATED:
            case E_USER_NOTICE:
            case E_USER_WARNING:
            case E_USER_DEPRECATED:
                $this->addError($errorArray);
                break;
            default:
                $errorResponse = new ErrorResponse();
                $errorResponse->setStatusCode(500);
                $errorResponse->setMessage($error_message);

                $error = $errorResponse->addNewEntity();
                $error->setMessage($error_message);
                $error->setFile($error_file);
                $error->setLine($error_line);
                //$error->setContext($error_context);

                $this->getResponseBuilder()->setResponse($errorResponse);
                $this->getResponseBuilder()->render();
        }

    }

    /**
     * @param ADataMiner $miner
     */
    protected function registerEndpoint($miner){
        $this->getMinerFactory()->addMiner($miner);
    }

    /**
     * @return MinerFactory
     */
    private function getMinerFactory(){
        if(is_null($this->minerFactory)){
            $this->minerFactory = new MinerFactory();
        }
        return $this->minerFactory;
    }

    /**
     * @return RequestBuilder
     */
    private function getRequestBuilder(){
        if(is_null($this->requestBuilder)){
            $this->requestBuilder = new RequestBuilder($this->getSerializer());
        }
        return $this->requestBuilder;
    }

    /**
     * @return ResponseBuilder
     */
    private function getResponseBuilder(){
        if(is_null($this->responseBuilder)){
            $this->responseBuilder = new ResponseBuilder($this->getSerializer());
        }
        return $this->responseBuilder;
    }

    /**
     * @return ISecurityHandler
     */
    private function getSecurityHandler(){
        return $this->securityHandler;
    }

    /**
     * @param ISecurityHandler $securityHandler
     * @return AImportEshopApi
     */
    private function setSecurityHandler($securityHandler){
        $this->securityHandler = $securityHandler;
        return $this;
    }

    /**
     * @return Serializer
     */
    private function getSerializer(){
        return $this->serializer;
    }

    /**
     * @param Serializer $serializer
     * @return AImportEshopApi
     */
    private function setSerializer($serializer){
        $this->serializer = $serializer;
        return $this;
    }

    /**
     * @return FakeRequest
     */
    protected function getFakeRequest(){
        if(is_null($this->fakeRequest)){
            $this->fakeRequest = new FakeRequest();
        }
        return $this->fakeRequest;
    }

    /**
     * @return FakeResponse
     */
    protected function getFakeResponse(){
        if(is_null($this->fakeResponse)){
            $this->fakeResponse = new FakeResponse();
        }
        return $this->fakeResponse;
    }

    /**
     * @return ADefaultEshopSetting
     * @throws MissingDefaultEshopSettingClassException
     */
    public function getDefaultEshopSetting(){
        if(is_null($this->defaultEshopSetting)){
            throw new MissingDefaultEshopSettingClassException();
        }
        return $this->defaultEshopSetting;
    }

    /**
     * @param ADefaultEshopSetting $defaultEshopSetting
     * @return AImportEshopApi
     */
    protected function setDefaultEshopSetting($defaultEshopSetting){
        $this->defaultEshopSetting = $defaultEshopSetting;
        return $this;
    }

    /**
     * @return array
     */
    public function getErrors(){
        return $this->errors;
    }

    /**
     * @param array $errors
     */
    public function setErrors($errors){
        $this->errors = $errors;
    }

    /**
     * @param string $message
     */
    private function addError($message){
        $this->errors[]=$message;
    }


}
