<?php

namespace Monkey\InsideAuth\Base;

use Monkey\Curl\Curl;
use Monkey\Curl\Exceptions\HttpCurlCommunicationException;
use Monkey\Curl\Exceptions\HttpCurlException;
use Monkey\Curl\Exceptions\HttpCurlUnauthorizedException;
use Monkey\InsideAuth\Exceptions\InsideAuthUnauthorizedException;
use Monkey\Oauth2Token\AccessToken;
use Monkey\Oauth2Token\Exceptions\Oauth2AccessTokenCreationException;
use Monkey\Oauth2Token\Interfaces\IAccessToken;
use stdClass;
use Throwable;
use Exception;
use Monkey\InsideAuth\Exceptions\InsideAuthCommunicationException;
use Monkey\InsideAuth\Exceptions\InsideAuthException;
use Monkey\InsideAuth\Interfaces\IInsideAuth;
use Monkey\InsidePlatformData\Base\AConfig;

/**
 * Class AInsideAuth
 * @package Monkey\InsideAuth\Base
 */
abstract class AInsideAuth implements IInsideAuth {

    /**
     * @var AConfig
     */
    private $config;

    /**
     * AInsideAuth constructor.
     * @param AConfig $config
     */
    public function __construct(AConfig $config) {
        $this->config = $config;
    }

    /**
     * @return AConfig
     */
    protected function getConfig() {
        return $this->config;
    }

    /**
     * @return string
     */
    private function getBaseUrl() {
        $parts = $this->getConfig()->getInsideUrlParts();
        $url = implode("/", $parts);
        return $url;
    }

    /**
     * @return string
     */
    private function getSecretTokenUrl() {
        return $this->getBaseUrl() . IInsideAuth::INSIDE_AUTH_SECRET_SLUG;
    }

    /**
     * @return string
     */
    private function getIframeTokenUrl() {
        return $this->getBaseUrl() . IInsideAuth::INSIDE_AUTH_IFRAME_SLUG;
    }

    /**
     * @param string $url
     * @param array $parameters
     * @return stdClass
     * @throws InsideAuthCommunicationException
     * @throws InsideAuthException
     */
    private function call($url, array $parameters) {
        try {
            $httpClient = new Curl();
            $response = $httpClient->call($url, $parameters);
            $data = $response->getData();

            if (!($data instanceof stdClass)) {
                throw new InsideAuthException("Invalid response data");
            }

            return $data;
        } catch (HttpCurlUnauthorizedException $exception){
            throw new InsideAuthUnauthorizedException($exception);

        } catch (HttpCurlCommunicationException $exception) {
            throw new InsideAuthCommunicationException($exception);

        } catch (HttpCurlException $exception) {
            throw new InsideAuthException($exception);

        } catch (Exception $exception) {
            throw new InsideAuthCommunicationException("Unknown communication error.", 0, $exception);

        } catch (Throwable $exception) {
            throw new InsideAuthCommunicationException("Unknown communication error.", 0, $exception);
        }

    }

    /**
     * @param string $url
     * @param array $parameters
     * @return IAccessToken
     * @throws InsideAuthException
     */
    private function getTokenFromInside($url, array $parameters) {
        $data = $this->call($url, $parameters);

        try {
            $token = AccessToken::make($data);
        } catch (Oauth2AccessTokenCreationException $exception) {
            throw new InsideAuthException("Token object creation error", 0, $exception);
        }

        return $token;
    }

    /**
     * @param string $authCode
     * @return IAccessToken
     * @throws InsideAuthException
     */
    public function exchangeAuthCodeForSecretToken($authCode) {
        $url = $this->getSecretTokenUrl();
        $parameters = array(IInsideAuth::FIELD_NAME_CODE => (string)$authCode);

        return $this->getTokenFromInside($url, $parameters);
    }

    /**
     * @param IAccessToken $secretToken
     * @return IAccessToken
     * @throws InsideAuthException
     */
    public function exchangeSecretTokenForIframeToken(IAccessToken $secretToken) {
        $url = $this->getIframeTokenUrl();
        $parameters = array(IInsideAuth::FIELD_NAME_ACCESS_TOKEN => $secretToken->getAccessToken());

        return $this->getTokenFromInside($url, $parameters);
    }

}
