<?php

namespace Monkey\Oauth2Token;

use Monkey\Oauth2Token\Base\AToken;
use Monkey\Oauth2Token\Exceptions\Oauth2AccessTokenCreationException;
use Monkey\Oauth2Token\Exceptions\Oauth2RefreshTokenCreationException;
use Monkey\Oauth2Token\Exceptions\Oauth2ScopeCreationException;
use Monkey\Oauth2Token\Interfaces\IAccessToken;
use Monkey\Oauth2Token\Interfaces\IRefreshToken;
use Exception;
use Monkey\Oauth2Token\Interfaces\IScope;
use stdClass;


/**
 * Class AccessToken
 * @package Monkey\Oauth2Token
 */
class AccessToken extends AToken implements IAccessToken {

    /**
     *
     * @var string
     */
    private $accessToken;

    /**
     * @var string
     */
    private $tokenType;

    /**
     * @var IRefreshToken
     */
    private $refreshToken;

    /**
     * @var IScope[]
     */
    private $scopes = array();

    /**
     * @return string
     */
    final public function getAccessToken() {
        return $this->accessToken;
    }

    /**
     * @param string $accessToken
     * @return $this
     */
    final public function setAccessToken($accessToken) {
        $this->accessToken = $accessToken;
        return $this;
    }

    /**
     * @return string
     */
    final public function getTokenType() {
        return $this->tokenType;
    }

    /**
     * @param string $tokenType
     * @return $this
     */
    final public function setTokenType($tokenType) {
        $this->tokenType = $tokenType;
        return $this;
    }

    /**
     * @return IRefreshToken|null
     */
    final public function getRefreshToken() {
        return $this->refreshToken;
    }

    /**
     * @param IRefreshToken $refreshToken
     * @return $this
     */
    final public function setRefreshToken(IRefreshToken $refreshToken) {
        $this->refreshToken = $refreshToken;
        return $this;
    }

    /**
     * @return bool
     */
    final public function hasRefreshToken() {
        if ($this->getRefreshToken() !== null) {
            return true;
        }
        return false;
    }

    /**
     * @return IScope[]
     */
    final public function getScopes() {
        return $this->scopes;
    }

    /**
     * @param string $code
     * @return bool
     */
    final public function hasScopeByCode($code) {
        if (isset($this->scopes[$code])) {
            return true;
        }
        return false;
    }

    /**
     * @param IScope[] $scopes
     * @return $this
     */
    final public function setScopes(array $scopes) {
        foreach ($scopes as $scope) {
            $this->addScope($scope);
        }
        return $this;
    }

    /**
     * @param IScope $scope
     * @return $this
     */
    final public function addScope(IScope $scope) {
        $this->scopes[$scope->getCode()] = $scope;
        return $this;
    }

    /**
     * @param stdClass $object
     * @return static
     * @throws Oauth2AccessTokenCreationException
     */
    public static function make(stdClass $object) {
        try {
            $accessToken = new static();


            if (isset($object->accessToken)) {
                $accessToken->setAccessToken($object->accessToken);
            }
            if (isset($object->{static::FIELD_NAME_ACCESS_TOKEN})) {
                $accessToken->setAccessToken($object->{static::FIELD_NAME_ACCESS_TOKEN});
            }

            if (isset($object->tokenType)) {
                $accessToken->setTokenType($object->tokenType);
            }
            if (isset($object->{static::FIELD_NAME_TOKEN_TYPE})) {
                $accessToken->setTokenType($object->{static::FIELD_NAME_TOKEN_TYPE});
            }

            if (isset($object->expiresAt)) {
                $accessToken->setExpiresAt($object->expiresAt);
            }
            if (isset($object->{static::FIELD_NAME_EXPIRES_AT})) {
                $accessToken->setExpiresAt($object->{static::FIELD_NAME_EXPIRES_AT});
            }

            if (isset($object->createdAt)) {
                $accessToken->setCreatedAt($object->createdAt);
            }
            if (isset($object->{static::FIELD_NAME_EXPIRES_IN})) {
                $createdAt = $accessToken->getExpiresAt()->getTimestamp() - $object->{static::FIELD_NAME_EXPIRES_IN};
                $accessToken->setCreatedAt($createdAt);
            }


            if (isset($object->refreshToken)) {
                try {
                    $refreshToken = RefreshToken::make($accessToken, $object->refreshToken);
                    $accessToken->setRefreshToken($refreshToken);
                } catch (Exception $exception) {
                    // refresh token creation error
                    throw new Oauth2RefreshTokenCreationException("Refresh token creation error", 0, $exception);
                }
            }

            if (isset($object->scopes)) {
                $scopes = (array)$object->scopes;

                foreach ($scopes as $scope) {
                    try {
                        $s = Scope::make($scope);
                        $accessToken->addScope($s);
                    } catch (Exception $exception) {
                        // scopes creation error
                        throw new Oauth2ScopeCreationException("Scope creation error", 0, $exception);
                    }
                }
            }

            return $accessToken;
        } catch (Exception $exception) {
            // access token creation error
            throw new Oauth2AccessTokenCreationException("Access token creation error", 0, $exception);
        }
    }

}