<?php

namespace Monkey\PluginsPrestashop16XApi\Endpoint;


use Exception;
use Monkey\ImportEshopApi\Endpoint\ACoupons;
use Monkey\ImportEshopApi\Exceptions\DataMiningFailedException;
use Monkey\ImportEshopDataObjects\Entity\Simple\Join;
use Monkey\ImportEshopDataObjects\Entity\Simple\Price;
use Monkey\PluginsPrestashop16XApi\Base\CouponRestrictionHelper;
use Monkey\PluginsPrestashop16XApi\Base\PrestaDefaultEshopSetting;
use Monkey\PluginsPrestashop16XApi\Facades\Db;
use Monkey\PluginsPrestashop16XApi\Facades\DbQuery;

/**
 * Class Coupons
 * @package Monkey\PluginsPrestashop16XApi\Endpoint
 */
class Coupons extends ACoupons {
    /**
     * @return string
     */
    protected function getComposedQuery() {
        $builder = new DbQuery();
        $this->getSelectAttributes($builder);
        $this->getTable($builder);
        $this->getConditions($builder);
        $this->getDefaultLimits($builder);
        $sql = $builder->build();
        return $sql;
    }

    /**
     * @return string
     * @throws DataMiningFailedException
     */
    protected function getTableSchema() {
        $tables = array("cart_rule","currency");

        $results = array();
        foreach($tables as $tableName){
            $tableName = $this->getDefaultEshopSetting()->getDbPrefix() . $tableName;
            $sql = "SHOW COLUMNS FROM {$tableName}";

            try {
                $columns = Db::executeS($sql);
            } catch (Exception $exception) {
                throw new DataMiningFailedException($exception);
            }

            foreach($columns as &$row){
                $row['table']=$tableName;
            }

            $results = array_merge($results, $columns);
        }

        return $results;
    }

    /**
     * @throws DataMiningFailedException
     */
    protected function setupResponseData() {
        $sql = $this->getComposedQuery();

        if ($this->getRequest()->isDebug()) {
            $this->getResponse()->setQuery($sql);
            $this->getResponse()->setSchema($this->getTableSchema());
        }

        try {
            $results = Db::executeS($sql);
        } catch (Exception $exception) {
            throw new DataMiningFailedException($exception);
        }

        if ($results) {
            $couponHelper = new CouponRestrictionHelper();

            foreach ($results as $row) {
                $minimumAmount = new Price();
                $minimumAmount->setPrice($row["order_minimum_amount"]);
                $minimumAmount->setCurrency($row["order_minimum_currency"]);

                $coupon = $this->getResponse()->addNewEntity();
                $coupon->setId($row["id"]);
                $coupon->setCode($row["code"]);
                $coupon->setAmount($row["amount"]);
                $coupon->setUsageCountLimit($row["usage_count_limit"]);
                $coupon->setOrderMinimumAmount($minimumAmount);
                $coupon->setStartDate($row["start_date"]);
                $coupon->setEndDate($row["end_date"]);
                $coupon->setCreatedAt($row["created_at"]);
                $coupon->setUpdatedAt($row["updated_at"]);
                $coupon->setIdShop($row["shop_id"]);

                if ($row["shipment"] == 1) {
                    $coupon->setShipmentFree();
                } else {
                    $coupon->setShipmentNone();
                }

                if ($row["type"] == "percentage") {
                    $coupon->setTypeToPercentage();
                } else {
                    $coupon->setTypeToAmount();
                }

                foreach ($couponHelper->getCategoriesRestrictions($coupon->getId()) as $categoryId) {
                    $coupon->addCategory($categoryId);
                }

                foreach ($couponHelper->getProductsRestrictions($coupon->getId()) as $productId) {
                    $coupon->addProduct($productId);
                }
            }
        }
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getDefaultSelectAttributes($builder = null) {
        $columns = array(
            "coupon.id_cart_rule AS id",
            "CONCAT(coupon_lang.name,' ',coupon.code) as code",
            "IF(coupon.reduction_percent > 0, 'percentage', 'amount') AS type",
            "IF(coupon.reduction_percent > 0, coupon.reduction_percent, coupon.reduction_amount) AS amount",
            "coupon.quantity AS usage_count_limit",
            "coupon.minimum_amount AS order_minimum_amount",
            "coupon_currency.iso_code AS order_minimum_currency",
            "coupon.free_shipping AS shipment",
            "coupon.date_from AS start_date",
            "coupon.date_to AS end_date",
            "coupon.date_add AS created_at",
            "coupon.date_upd AS updated_at",
            "coupon_shop.id_shop AS shop_id"
        );
        $builder->select(implode(", ", $columns));
        return true;
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getDefaultTable($builder = null) {
        $builder->from("cart_rule", "coupon");
        $builder->innerJoin("cart_rule_lang", "coupon_lang", "coupon.id_cart_rule = coupon_lang.id_cart_rule AND id_lang = '" . cSQL($this->getDefaultEshopSetting()->getLanguageId()) . "'");
        $builder->leftJoin("currency", "coupon_currency", "coupon.minimum_amount_currency = coupon_currency.id_currency");
        return true;
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getDefaultCondition($builder = null) {
        $shopId = $this->getRequest()->getCurrentDefaultSetting("id_shop");

        if (!is_null($shopId)) {
            $builder->leftJoin("cart_rule_shop", "coupon_shop", "coupon.id_cart_rule = coupon_shop.id_cart_rule AND coupon_shop.id_shop = '" . cSQL($shopId->getValue()) . "'");
        }

        if (!is_null($this->getRequest()->getId())) {
            $builder->where("coupon.id_cart_rule = '" . cSQL($this->getRequest()->getId()) . "'");
            return true;
        }

        //Select by dates
        $createdCondition = "";
        $updatedCondition = "";
        $dateCondition = "";
        $createdFrom = $this->getRequest()->getCreatedFrom();
        $createdTo = $this->getRequest()->getCreatedTo();
        $updatedFrom = $this->getRequest()->getUpdatedFrom();
        $updatedTo = $this->getRequest()->getUpdatedTo();

        if (!is_null($createdFrom) && !is_null($createdTo)) {
            $createdCondition .= "coupon.date_add BETWEEN '" . cSQL($createdFrom) . "' AND '" . cSQL($createdTo) . "'";
        } else if (!is_null($createdFrom)) {
            $createdCondition .= "coupon.date_add >= '" . cSQL($createdFrom) . "'";
        } else if (!is_null($createdTo)) {
            $createdCondition .= "coupon.date_add <= '" . cSQL($createdTo) . "'";
        }

        if (!is_null($updatedFrom) && !is_null($updatedTo)) {
            $updatedCondition .= "coupon.date_upd BETWEEN '" . cSQL($updatedFrom) . "' AND '" . cSQL($updatedTo) . "'";
        } else if (!is_null($updatedFrom)) {
            $updatedCondition .= "coupon.date_upd >= '" . cSQL($updatedFrom) . "'";
        } else if (!is_null($updatedTo)) {
            $updatedCondition .= "coupon.date_upd <= '" . cSQL($updatedTo) . "'";
        }

        if (!empty($createdCondition) && !empty($updatedCondition)) {
            $dateCondition = $createdCondition . " OR " . $updatedCondition;
        } else {
            $dateCondition = $createdCondition . $updatedCondition;
        }

        if (!empty($dateCondition)) {
            $builder->where($dateCondition);
            return true;
        }

        return false;
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getDefaultLimits($builder = null) {
        $builder->limit($this->getResultsCount(), $this->getOffset());
        return true;
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getDifferentAttributes($builder = null) {
        $return = false;

        foreach ($this->getRequest()->getDifferences()->getColumns() as $column) {
            $return = true;
            $builder->select($column->__toString());
        }

        if ($return) return true;
        return null;
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getDifferentTable($builder = null) {
        $return = false;

        foreach ($this->getRequest()->getDifferences()->getTableAliases() as $table) {
            $return = true;

            if ($table->hasConditions()) {
                $conditions = "";

                foreach ($table->getConditions() as $condition) {
                    $conditions .= $condition->__toString();
                }

                $conditions = $this->getRequest()->setupVariables($conditions);

                switch ($table->getType()) {
                    case Join::INNER:
                        $builder->innerJoin($table->getTableName(), $table->getAlias(), $conditions);
                        break;
                    case Join::LEFT:
                        $builder->leftJoin($table->getTableName(), $table->getAlias(), $conditions);
                        break;
                }
            } else {
                $builder->from($table->getTableName(), $table->getAlias());
            }
        }

        if ($return) return true;
        return null;
    }

    /**
     * @param mixed|null $builder
     * @return mixed|null
     */
    protected function getDifferentConditions($builder = null) {
        $return = false;
        $conditions = "";

        foreach ($this->getRequest()->getDifferences()->getAdditionalConditions() as $condition) {
            $return = true;
            $conditions .= $condition->__toString();
        }

        $builder->where($this->getRequest()->setupVariables($conditions));

        if ($return) return true;
        return null;
    }

    /**
     * @return PrestaDefaultEshopSetting
     */
    protected function getDefaultEshopSetting() {
        return parent::getDefaultEshopSetting();
    }
}