<?php

namespace Monkey\PluginsPrestashop16XApi\Endpoint;


use Exception;
use Monkey\ImportEshopApi\Endpoint\AOrders;
use Monkey\ImportEshopApi\Exceptions\DataMiningFailedException;
use Monkey\ImportEshopDataObjects\Entity\Customer;
use Monkey\ImportEshopDataObjects\Entity\Simple\Discount;
use Monkey\ImportEshopDataObjects\Entity\Simple\Join;
use Monkey\ImportEshopDataObjects\Entity\Simple\Payment;
use Monkey\ImportEshopDataObjects\Entity\Simple\Price;
use Monkey\ImportEshopDataObjects\Entity\Simple\Shipment;
use Monkey\PluginsPrestashop16XApi\Base\PrestaDefaultEshopSetting;
use Monkey\PluginsPrestashop16XApi\Facades\Db;
use Monkey\PluginsPrestashop16XApi\Facades\DbQuery;

/**
 * Class Orders
 * @package Monkey\PluginsPrestashop16XApi\Endpoint
 */
class Orders extends AOrders {
    /**
     * @param mixed|null $builder
     * @return string
     */
    protected function getComposedQuery($builder = null) {
        $builder = new DbQuery();
        $this->getSelectAttributes($builder);
        $this->getTable($builder);
        $this->getConditions($builder);
        $this->getDefaultLimits($builder);
        $builder->groupBy("o.id_order");
        $sql = $builder->build();
        return $sql;
    }

    /**
     * @throws DataMiningFailedException
     */
    protected function setupResponseData() {
        $sql = $this->getComposedQuery();

        try {
            $results = Db::executeS($sql);
        } catch (Exception $exception) {
            throw new DataMiningFailedException($exception);
        }

        if ($results) {
            $orderIds = array();

            foreach ($results as $row) {
                $orderIds[] = $row['id'];
            }

            $couponsBuilder = new DbQuery();
            $couponsBuilder->select("`ocr`.*, `cr`.`code`");
            $couponsBuilder->from("order_cart_rule", "ocr");
            $couponsBuilder->leftJoin("cart_rule","cr","cr.id_cart_rule = ocr.id_cart_rule");
            $couponsBuilder->where("id_order IN (".cSQL(implode(", ", $orderIds).")"));
            try{
                $coupons = Db::executeS($couponsBuilder->build());
            }catch(\Exception $exception){
                throw new DataMiningFailedException($exception);
            }

            $discounts = array();
            if($coupons) {
                foreach ($coupons as $coupon) {
                    $discounts[$coupon['id_order']][$coupon['id_order_cart_rule']] = $coupon;
                }
            }

            foreach ($results as $row) {
                $order = $this->getResponse()->addNewEntity();
                $order->setId($row['id']);
                $order->setIdShop($row["id_shop"]);
                $order->setIdShopGroup($row["id_shop_group"]);
                $order->setVisibleOrderId($row["id"]);
                $order->setSourceType(self::ESHOP_SOURCE);
                $order->setNote($row["note"]);
                $order->setCreatedAt($row['created']);
                $order->setUpdatedAt($row['updated']);
                $order->setOrderStatusId($row['order_status_id']);
                $order->setOrderStatusName($row['order_status_name']);

                $freeShipping = false;

                if (isset($discounts[$row['id']])) {
                    foreach ($discounts[$row['id']] as $coupon) {
                        $discount = new Discount();
                        $discount->setId($coupon['id_cart_rule']);
                        $discount->setName($coupon['code']);
                        $discount->setValue($coupon['value']);

                        if ($coupon['free_shipping'] == 1) {
                            $freeShipping = true;
                            $discount->setValue(0);
                            $discount->setType("free_shipping");
                        }
                        $order->addDiscount($discount);
                    }
                }

                $priceShipment = new Price();
                $priceShipment->setPrice($row['shipping_price']);
                $priceShipment->setPriceWithoutVat($row['shipping_price_without_vat']);
                $priceShipment->setCurrency($row["iso_code"]);
                $priceShipment->setRate($row['carrier_tax_rate']);
                $priceShipment->setConversionRate($row["conversion_rate"]);

                if ($freeShipping) {
                    $priceShipment->setPrice(0);
                    $priceShipment->setPriceWithoutVat(0);
                }

                $shipment = new Shipment();
                $shipment->setName($row['shipping_name']);
                $shipment->setPrice($priceShipment);

                $pricePayment = new Price();
                $pricePayment->setPrice(0);
                $pricePayment->setCurrency($row["iso_code"]);

                $payment = new Payment();
                $payment->setName($row['payment_name']);
                $payment->setPrice($pricePayment);

                $price = new Price();
                $price->setCurrency($row["iso_code"]);
                $price->setPrice($row['price']);
                $price->setPriceWithoutVat($row['price_without_vat']);
                $price->setConversionRate($row['conversion_rate']);

                $customer = new Customer();
                $customer->setId($row['customer_id']);
                $customer->setEmail($row['email']);
                $customer->setFirstname($row['firstname']);
                $customer->setLastname($row['lastname']);
                $customer->setCity($row['city']);
                $customer->setCountry($row['country']);
                $customer->setZipCode($row['postcode']);
                $customer->setRegistered(intval($row['not_registered']) ? false : true);
                $customer->setCompany(!is_null($row['company']));
                $customer->setVatPayer(!is_null($row['vat_number']));

                $order->setShipment($shipment);
                $order->setPayment($payment);
                $order->setCustomer($customer);
                $order->setPrice($price);
            }
        }
    }

    /**
     * @param DbQuery|null $builder
     * @return mixed
     */
    protected function getDefaultSelectAttributes($builder = null) {
        $columns = array(
            "o.id_order as id",
            "o.`date_add` as created",
            "o.`date_upd` as updated",
            "o.total_paid_tax_incl as price",
            "o.total_paid_tax_excl as price_without_vat",
            "o.carrier_tax_rate as rate",
            "o.current_state as order_status_id",
            "osl.name as order_status_name",
            "o.id_shop","o.id_shop_group",
            "o.conversion_rate",
            "op.payment_method as payment_name",
            "op.amount as payment_price",
            "ca.name as shipping_name",
            "o.total_shipping_tax_incl as shipping_price",
            "o.total_shipping_tax_excl as shipping_price_without_vat",
            "c.id_customer as customer_id",
            "c.firstname",
            "c.lastname",
            "c.email",
            "c.is_guest as not_registered",
            "c.company",
            "a.postcode",
            "a.city",
            "a.vat_number",
            "cl.name as country",
            "cu.iso_code",
            "GROUP_CONCAT(m.message) as note");
        $builder->select(implode(",",$columns));
        return true;
    }

    /**
     * @param DbQuery|null $builder
     * @return mixed
     */
    protected function getDefaultTable($builder = null) {
        $builder->from("orders","o");
        $builder->leftJoin("customer","c","o.id_customer = c.id_customer");
        $builder->leftJoin("currency","cu","o.id_currency = cu.id_currency");
        $builder->leftJoin("order_payment","op","o.reference = op.order_reference");
        $builder->leftJoin("order_state_lang","osl","osl.id_order_state = o.current_state AND osl.id_lang = ".cSQL($this->getDefaultEshopSetting()->getLanguageId()));
        $builder->leftJoin("address","a","o.id_address_delivery = a.id_address");
        $builder->leftJoin("country_lang","cl","a.id_country = cl.id_country AND cl.id_lang = ".cSQL($this->getDefaultEshopSetting()->getLanguageId()));
        $builder->leftJoin("order_carrier","oc","oc.id_order = o.id_order");
        $builder->leftJoin("carrier","ca","ca.id_carrier = oc.id_carrier");
        $builder->leftJoin("message", "m", "m.id_order = o.id_order");

        return true;
    }

    /**
     * @param DbQuery|null $builder
     * @return mixed
     */
    protected function getDefaultCondition($builder = null) {
        $shopId = $this->getRequest()->getCurrentDefaultSetting("id_shop");
        $shopGroupId = $this->getRequest()->getCurrentDefaultSetting("id_shop_group");

        if (!is_null($shopId)) {
            $builder->where("o.id_shop = '" . cSQL($shopId->getValue()) . "'");
        }

        if (!is_null($shopGroupId)) {
            $builder->where("o.id_shop_group = '" . cSQL($shopGroupId->getValue()) . "'");
        }

        if (!is_null($this->getRequest()->getId())) {
            $builder->where("o.id_order = ".cSQL($this->getRequest()->getId()));
            return true;
        }

        //Select by dates
        $createdCondition = "";
        $updatedCondition = "";
        $dateCondition = "";
        $createdFrom = $this->getRequest()->getCreatedFrom();
        $createdTo = $this->getRequest()->getCreatedTo();
        $updatedFrom = $this->getRequest()->getUpdatedFrom();
        $updatedTo = $this->getRequest()->getUpdatedTo();

        if (!is_null($createdFrom) && !is_null($createdTo)) {
            $createdCondition .= "o.date_add BETWEEN '" . cSQL($createdFrom) . "' AND '" . cSQL($createdTo) . "'";
        } else if (!is_null($createdFrom)) {
            $createdCondition .= "o.date_add >= '" . cSQL($createdFrom) . "'";
        } else if (!is_null($createdTo)) {
            $createdCondition .= "o.date_add <= '" . cSQL($createdTo) . "'";
        }

        if (!is_null($updatedFrom) && !is_null($updatedTo)) {
            $updatedCondition .= "o.date_upd BETWEEN '" . cSQL($updatedFrom) . "' AND '" . cSQL($updatedTo) . "'";
        } else if (!is_null($updatedFrom)) {
            $updatedCondition .= "o.date_upd >= '" . cSQL($updatedFrom) . "'";
        } else if (!is_null($updatedTo)) {
            $updatedCondition .= "o.date_upd <= '" . cSQL($updatedTo) . "'";
        }
        if (!empty($createdCondition) && !empty($updatedCondition)) {
            $dateCondition = $createdCondition . " OR " . $updatedCondition;
        } else {
            $dateCondition = $createdCondition . $updatedCondition;
        }

        if (!empty($dateCondition)) {
            $builder->where($dateCondition);
            return true;
        }

        return false;
    }

    /**
     * @param DbQuery|null $builder
     * @return mixed
     */
    protected function getDefaultLimits($builder = null){
        $builder->limit($this->getResultsCount(), $this->getOffset());
        return true;
    }


    /**
     * fce zpracuje prijate column differences z requestu pro konkretni implementaci platformy
     *
     * @param DbQuery|null $builder
     * @return mixed|null
     */
    protected function getDifferentAttributes($builder = null) {
        $return = false;

        foreach ($this->getRequest()->getDifferences()->getColumns() as $column) {
            $return = true;
            $builder->select($column->__toString());
        }

        if ($return) return true;
        return null;
    }

    /**
     * @param DbQuery|null $builder
     * @return mixed|null
     */
    protected function getDifferentTable($builder = null) {
        $return = false;

        foreach ($this->getRequest()->getDifferences()->getTableAliases() as $table) {
            $return = true;

            if ($table->hasConditions()) {
                $conditions = "";

                foreach ($table->getConditions() as $condition) {
                    $conditions .= $condition->__toString();
                }

                $conditions = $this->getRequest()->setupVariables($conditions);

                switch ($table->getType()) {
                    case Join::INNER:
                        $builder->innerJoin($table->getTableName(), $table->getAlias(), $conditions);
                        break;
                    case Join::LEFT:
                        $builder->leftJoin($table->getTableName(), $table->getAlias(), $conditions);
                        break;
                }
            } else {
                $builder->from($table->getTableName(), $table->getAlias());
            }
        }

        if ($return) return true;
        return null;
    }

    /**
     * @param DbQuery|null $builder
     * @return mixed|null
     */
    protected function getDifferentConditions($builder = null) {
        $return = false;
        $conditions = "";

        foreach ($this->getRequest()->getDifferences()->getAdditionalConditions() as $condition) {
            $return = true;
            $conditions .= $condition->__toString();
        }

        $builder->where($this->getRequest()->setupVariables($conditions));

        if ($return) return true;
        return null;
    }

    /**
     * @return string
     * @throws DataMiningFailedException
     */
    protected function getTableSchema(){
        $tables = array("orders","customer","currency","order_payment","order_state_lang","address","country_lang","order_carrier","carrier","message");
        $results = array();
        foreach($tables as $tableName){
            $tableName = $this->getDefaultEshopSetting()->getDbPrefix() . $tableName;
            $sql = "SHOW COLUMNS FROM {$tableName}";

            try {
                $columns = Db::executeS($sql);
            } catch (Exception $exception) {
                throw new DataMiningFailedException($exception);
            }

            foreach($columns as &$row){
                $row['table']=$tableName;
            }

            $results = array_merge($results, $columns);
        }

        return $results;
    }

    /**
     * @return PrestaDefaultEshopSetting
     */
    protected function getDefaultEshopSetting(){
        return parent::getDefaultEshopSetting();
    }

}